"""
Global constants for the Perception module.
"""
METERS_TO_MM = 1000.0
MM_TO_METERS = 1.0 / METERS_TO_MM

COLOR_IMAGE_EXTS = ['.png', '.jpg']
MIN_DEPTH = 0.25
MAX_DEPTH = 1.25
MAX_IR = 65535

TF_EXTENSION = '.tf'
INTR_EXTENSION = '.intr'
